// VinciaBranchelemental.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2013 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Branchelemental class 
  
#ifndef VINCIA_VinciaBranchelemental_H
#define VINCIA_VinciaBranchelemental_H

#include "Pythia8/Event.h"
#include "VinciaTrialGenerators.h"

using namespace Pythia8;

//**************************************************************************

// The Branchelemental class: contains a generic 2->3 branching, with 
// pointers to the relevant places in the event record and a few useful 
// functions. 
// The member function list() prints a brief record of the contents.
// The member function setTrial(bool) is used to (re)set whether this
// elemental already contains a saved trial branching, or whether a new 
// one should be generated.

namespace Vincia {

  class Branchelemental {
    
  public:
    
    // Constructors
    // Create empty branchelemental
    Branchelemental() : system(0) {}
    // Create branchelemental with iOld1In,iOld2In as parents
    // and initialize the PDT pointers for the new ones
    Branchelemental(Event& event, int iOld1In, int iOld2In)
    : system(0) {reset(event, iOld1In, iOld2In);};
    
    // Antenna Mass (negative if spacelike virtual)
    double mCalc() {
      sSav = m2(old1,old2);
      mSav = (sSav >= 0) ? sqrt(sSav) : -sqrt(-sSav);
      return mSav;
    }
    double m() const { return mSav; }
    double s() const { return sSav; }
    // CM energy squared of daughters (for cross checks)
    double sNew() const {return m2(new1,new2,new3);}
    // Dot products of daughters
    double s12() const {return m2(new1,new2) - new1.m2() - new2.m2();} 
    double s23() const {return m2(new2,new3) - new2.m2() - new3.m2();} 
    double s13() const {return m2(new1,new3) - new1.m2() - new3.m2();} 
    double sNew1() const {return new1.m2();} 
    double sNew2() const {return new2.m2();}
    double sNew3() const {return new3.m2();}

    // Reset and set trial true/false
    void reset(Event& event, int iOld1In, int iOld2In) {
      // Make local copy of input particles
      old1        = event[iOld1In];
      old2        = event[iOld2In];
      // Compute and store antenna invariant mass
      mCalc();      
      // Indices of parents and neighbors
      iOld1       = iOld1In;
      iOld2       = iOld2In;
      iL          = 0; 
      iR          = 0;
      iLL         = 0;
      iRR         = 0;
      // Trial Generators
      clearTrialGenerators();
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
      // Default antenna properties
      Qold        = 0.0;
      unOrdering  = false;
      new1=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      new2=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      new3=Particle(0,51,iOld2,iOld2,0,0,0,0,0.);      
      // Set pointers
      new1.setEvtPtr(&event);
      new2.setEvtPtr(&event);
      new3.setEvtPtr(&event);
    }

    // Function to reset all trial generators for this branchelemental
    void clearTrialGenerators() {
      trialGenPtrSav.resize(0);
      iAntPhysSav.resize(0);
      isSwappedSav.resize(0);
      hasSavedTrial.resize(0);
      scaleSav.resize(0);
      zMinSav.resize(0);
      zMaxSav.resize(0);
      colFacSav.resize(0);
      alphaSav.resize(0);
      headRoomSav.resize(0);
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
    }

    // Add a trial generator to this branchelemental. Return name of generator.
    int addTrialGenerator(int iAntPhysIn, bool swapIn,  
                          TrialGenerator* trialGenPtrIn) {
      trialGenPtrSav.push_back(trialGenPtrIn);
      iAntPhysSav.push_back(iAntPhysIn);
      isSwappedSav.push_back(swapIn);
      hasSavedTrial.push_back(false);      
      scaleSav.push_back(-1.0);
      zMinSav.push_back(0.0);
      zMaxSav.push_back(0.0);
      colFacSav.push_back(0.0);
      alphaSav.push_back(0.0);
      headRoomSav.push_back(1.0);
      return int(trialGenPtrSav.size()-1);
    }
    
    // Function to return number of trial generators for this antenna
    int nTrialGenerators() const { return trialGenPtrSav.size(); }

    // Save a generated trial branching
    void saveTrial(int iTrial,double qTrial,double zMin,double zMax,
                   double colFac,double alphaEff, double headRoom = 1.0) {
      hasSavedTrial[iTrial] = true;
      scaleSav[iTrial]      = qTrial;
      zMinSav[iTrial]       = zMin;
      zMaxSav[iTrial]       = zMax;
      colFacSav[iTrial]     = colFac;
      alphaSav[iTrial]      = alphaEff;
      headRoomSav[iTrial]   = headRoom;
    }

    // Generate invariants for saved branching
    bool genTrialInvariants(double& y12, double& y23, int iTrial = -1) {
      // Automatically determine which trial function to use if -1 input
      int iGen = iTrial;
      if (iGen == -1) iGen = getTrialIndex();
      if (iGen <= -1) return false; 
      double z = trialGenPtrSav[iGen]->genZ(zMinSav[iGen],zMaxSav[iGen]);
      // Check physical phase space (note: this only checks massless hull)
      // (Use absolute z value since negative z values are used to indicate 
      //  swapped invariants for mD ordering)
      double yE = pow2(scaleSav[iGen]/mSav);
      if (abs(z) < trialGenPtrSav[iGen]->getZmin(yE) || 
          abs(z) > trialGenPtrSav[iGen]->getZmax(yE)) return false;
      // Convert to y12 , y23
      y12 = trialGenPtrSav[iGen]->getYij(yE,z);
      y23 = trialGenPtrSav[iGen]->getYjk(yE,z);
      return true;
    }

    // Get trial function index of winner
    // (return -1 if not all trial generators have saved trials)
    int getTrialIndex() const {
      double qMax = 0.0;
      int iMax    = -1;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (!hasSavedTrial[i]) return -1;
        double qSav = scaleSav[i];
        if (qSav > qMax) {
          qMax = qSav;
          iMax = i;
        }
      }
      return iMax;
    }
        
    // Check if a saved trial exists for a particular trialGenerator
    bool hasTrial(int iTrial) const {
      if (iTrial < int(hasSavedTrial.size())) return hasSavedTrial[iTrial];
      else return false;
    }

    // Get whether physical antenna is swapped 
    bool getIsSwapped(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return isSwappedSav[iTrial];
    }
    
    // Get physical antenna function index of winner
    int getPhysIndex(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return iAntPhysSav[iTrial];
    }
    
    // Get scale for a specific saved trial
    double getTrialScale(int iTrial) const {
      if (iTrial < int(scaleSav.size())) return scaleSav[iTrial];
      else return -1.0;
    }

    // Get scale of winner
    double getTrialScale() const {
      double qMax = 0.0;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (hasSavedTrial[i]) qMax = max(qMax,scaleSav[i]);
        else printOut("getTrialScale","Error! "
                      "not all trials have saved scales");
      }
      return qMax;
    }

    // Get color factor 
    double getColFac(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return colFacSav[iTrial];
    }
    
    // Get whether antenna is swapped 
    double getHeadRoomFactor(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return headRoomSav[iTrial];
    }
    
    // Get whether antenna is swapped 
    double getAlphaTrial(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return alphaSav[iTrial];
    }
    
    // Flag to set if a saved trial should be ignored and a new one generated
    // Default value -1 : force all trials to renew
    void renewTrial(int iTrial = -1) {
      if (iTrial >= 0) hasSavedTrial[iTrial] = false;
      else {
        for (iTrial = 0; iTrial < int(hasSavedTrial.size()); ++iTrial)
          hasSavedTrial[iTrial] = false;
      }
    }

    // List function
    void list(ostream& os=cout, bool header=false, bool footer=false) const; 
    
    // Data storage members 
    int iOld1, iOld2;    
    Particle old1, old2;
    Particle new1, new2, new3;  
    // Nearest-Neighbour and Next-to-Nearest Neighbour indices
    int iL,iR;
    int iLL,iRR;
    // 
    int system,nVeto,nHull,nHadr;
    double Qold;
    // Ultimate ancestor (e.g., if part of decay system)
    int iAncestor;
    // Has trial branching, and whether strong ordering should be imposed
    bool unOrdering;  

    // Trial Generators and properties of saved trials
    vector<TrialGenerator*> trialGenPtrSav;
    vector<double> zMinSav, zMaxSav, colFacSav, alphaSav;
    vector<double> scaleSav, headRoomSav;
    vector<bool> hasSavedTrial, isSwappedSav;
    vector<int> iAntPhysSav;

   private:

    // Saved antenna invariant mass value
    double mSav, sSav;

  };
  
}

#endif
