// vincia10-root.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study energy scaling of ee observables
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2013 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
#include "PythiaRoot.h"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

static const int nEnergies = 14;
static const double eCM[nEnergies] = {14.,22.,29.,34.8,43.6,57.8,
                                      91.2,91.2,91.2,91.2,
                                      133.,161.,183.,200.};

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects. 
  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia10.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event* event = &pythia.event;

  // Some useful pointers
  Settings* settingsPtr   = vincia.settingsPtr;        // to settings database  

  // Extract settings to be used in the main program.  
  int    nEvent           = settingsPtr->mode("Main:numberOfEvents");  

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  int    verbose          = settingsPtr->mode("Vincia:verbose");

  // Matching settings
  int    matchingLO       = settingsPtr->mode("Vincia:matchingLO");
  bool   matching         = matchingLO >= 1;

  // hadronization settings
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");

  //************************************************************************
  // VinciaRoot
  // Create VinciaRoot object and set plot labels, legends, etc
  // Also define names of files to save and read histograms to/from
  //************************************************************************

  // Create VinciaRoot object and set default behavior to logarithmic y axes
  VinciaRoot vroot; 
  vroot.setLogY(false);
  // Set range on ratio plots (default is [0.5,1.5])
  vroot.setRatioRange(0.75,1.25);

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose(verbose);
  
  // Set first update and save event
  double iSave   = 256;  

  // Set MC names and versions
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  string mcVersions = "Vincia " + vcVer 
    + (matching ? " + MadGraph 4.4.26" : "") + " + Pythia "+pyVer;

  // Examples of how to set plot information and define read/save file names
  string mcLabel;
  // Example 1: PYTHIA runs
  if (!vinciaOn) {
    mcLabel    = "Pythia";
    mcVersions = "Pythia "+pyVer;
  }
  // Example 2: VINCIA runs
  else {
    mcLabel    = "Vincia";
  }
  // Example 3: parton-level results (hadronization OFF)
  if (!hadronLevel) mcLabel += " partons";
  
  // Save results of current run in data files with the following prefix
  string mcFileSave = "validation/vincia-def/vincia10";

  // Compare current run to files with the following prefix(es)
  vector<string> mcFileRead;
  mcFileRead.push_back("validation/pythia-def/vincia10");
  mcFileRead.push_back("validation/vincia-def/vincia10");

  // Set VinciaRoot plot labels, legends, etc
  vroot.setLegend(mcLabel);
  vroot.setMCnames(mcVersions);

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythia.init();

  //************************************************************************
  // Book Histograms
  //************************************************************************

  // Default plot sizes
  double plotSizeX = 360;
  double plotSizeY = 420;

  // Histogram bin and range variables
  int nBins;
  double xLo, xHi;

  // 
  nBins = nEnergies;
  xLo   = -0.5;
  xHi   = xLo + nEnergies;
  vector<TH1D*> VNch;
  VNch.push_back(new TH1D("Nch","Average Charged Multiplicity vs ECM",
                          nBins, xLo, xHi));
  VNch[0]->GetYaxis()->SetTitle("<n_{ch}>");
  VNch[0]->GetXaxis()->SetTitle("E_{cm} (not to scale)");
  for (int iBin=1; iBin<=nBins; ++iBin) 
    VNch[0]->GetXaxis()->SetBinLabel(iBin,string(num2str(int(eCM[iBin-1]+0.49))).c_str());
  VNch[0]->SetMaximum(40.01); 
  VNch[0]->SetMinimum(-0.8); 
  VNch[0]->GetXaxis()->SetLabelOffset(0.01);
  VNch[0]->GetXaxis()->SetLabelSize(0.05);
  VNch[0]->GetXaxis()->SetTitleOffset(1.75);
  vroot.createDisplay("Nch",VNch,plotSizeX,plotSizeY);
  vroot.setNormalize("Nch",false);
  vroot.setDataFile("Nch","data/ee-All-Nch.vec",25,1,40,50,60,70);  
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("Nch",mcFileRead[iRead]+"-Nch.dat");
  
  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // Loop over CM energies
  double eCMold = 0.0;
  for (int iECM=0; iECM<nEnergies; ++iECM) {
    double eCMnow = eCM[iECM];
    // Skip if several identical energies in succession
    if (eCMnow == eCMold) continue;
    eCMold = eCMnow;
    
    // Initialize PYTHIA for this CM energy
    pythia.readString("Beams:eCM = "+num2str(eCMnow));
    pythia.init();
    
    // Begin event loop
    double nCsum    = 0.0;
    double nC2sum   = 0.0;
    double iSaveNow = iSave;
    for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

      // Generate next event
      if (!pythia.next()) continue;
      event = &pythia.event;
      
      // Count up average charged particles
      double nC = 0;
      for (int i=1;i<event->size();i++) {
        if ((*event)[i].isCharged() && (*event)[i].status() > 80) ++nC; 
      }
      nCsum  += nC;
      nC2sum += pow2(nC);

      // Periodically update displays
      if ( iEvent == int(iSaveNow) || 
           (iEvent+251)%100 == 0 || iEvent == nEvent-1) {
        // Fill all bins that matches this CM energy
        for (int jECM=0; jECM<nEnergies; ++jECM) {
          if (eCM[jECM] != eCMnow) continue;
          double avg  = nCsum/(iEvent+1);
          double rms  = sqrt(nC2sum/iEvent+1);
          double sig2 = max(0.,pow2(rms)-pow2(avg));
          double err  = sqrt(sig2/(iEvent+1));
          VNch[0]->SetBinContent(jECM+1,avg);
          VNch[0]->SetBinError(jECM+1,err);
        }
        vroot.updateDisplays();       
      }
      
      // Periodically save displays
      if (iEvent == int(iSaveNow) || iEvent == nEvent-1) {
        iSaveNow *= sqrt(2.0);
        iSaveNow = min(iSaveNow,double(nEvent-1));
        cout<<"vincia10-root: saving plots at event "<<int(iEvent);
        cout<<" next save at event "<<int(iSaveNow)<<endl;
        // Save displays with base name mcFile and file type .eps
        vroot.saveDisplays(mcFileSave,"pdf");
        // Also save histograms as dat files
        vroot.saveHistograms(mcFileSave,"dat");
        // Print chi2 values
        // vroot.printChi2();
        // Print stats
      }
    }

    // Print out Vincia internal histograms
    if (vinciaOn && verbose >= 2) vincia.printHistos();
    // Print out Vincia end-of-run information
    pythia.statistics();
    if (vinciaOn) vincia.printInfo();

  } // end loop over CM energies
  
  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
